class
	AST_OCTAL_CONSTANT

inherit
	AST_EXPRESSION
		rename
			type_binding as type
		end

create
	make_from_antlr,
	make_copy

feature {NONE}

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_octal_constant
		local
			s: STRING
		do
			parent := init_parent
			s := node.children.first.token_text.as_lower
			if s.ends_with ("ull") then
				value := s.substring (1, s.count - 3)
				type := create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_long_long)
			elseif s.ends_with ("ul") then
				value := s.substring (1, s.count - 2)
				type := create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_long)
			elseif s.ends_with ("ll") then
				value := s.substring (1, s.count - 2)
				type := create {AST_BASIC_TYPE}.make (Ast_bt_long_long)
			elseif s.ends_with ("l") then
				value := s.substring (1, s.count - 1)
				type := create {AST_BASIC_TYPE}.make (Ast_bt_long)
			elseif s.ends_with ("u") then
				value := s.substring (1, s.count - 1)
				type := create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
			else
				value := s
				type := create {AST_BASIC_TYPE}.make (Ast_bt_int)
			end
			type.parent := Current
		end

feature {ANY}

	value: STRING

	type: AST_BASIC_TYPE

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_octal_constant (Current)
			if visit_children then
				visit_node (type, v)
			end
			v.end_visit_octal_constant (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			value := other.value.twin
			type := other.type.ast_twin
			type.parent := Current
		end

invariant
	value /= Void and then value.starts_with ("0")
	type /= Void and then
			(type.type = Ast_bt_int or type.type = Ast_bt_long or type.type = Ast_bt_long_long or
			type.type = Ast_bt_unsigned_int or type.type = Ast_bt_unsigned_long or type.type = Ast_bt_unsigned_long_long)

end
