class
	ADDRESS_OPERATOR_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_unary_expression
		end
	CLASS_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Adapting the %"address of%" operator where necessary"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_unary_expression (node: AST_UNARY_EXPRESSION)
		local
			l_cur: AST_POSTFIX_EXPRESSION
			l_is_struct_field_address: BOOLEAN
		do
			if node.type = C_op_bitand then
				if
					-- `&foo[bar]'  ->  `foo + bar'
					attached {AST_POSTFIX_EXPRESSION} node.expression as pe and then
					pe.type = C_op_open_square_bracket
				then
					from
						l_cur ?= pe.main_expression.non_parentheses_expression
					until
						l_is_struct_field_address or l_cur = Void
					loop
						if l_cur.type = C_op_arrow then
							l_is_struct_field_address := True
						end
						l_cur ?= l_cur.main_expression.non_parentheses_expression
					end
					if l_is_struct_field_address then
						handle_struct_field_address (node)
					else
						replace_node (node, create {AST_GENERIC_EXPRESSION}.make (
								C_op_plus,
								pe.main_expression.ast_twin,
								pe.array_expression.ast_twin
							))
					end
				elseif
					-- `&foo'  ->  `foo_address'  where foo is a function
					attached {AST_VARIABLE_EXPRESSION} node.expression as variable_expression and then
					attached {AST_FUNCTION_DECLARATION} variable_expression.declaration_binding as declaration
				then
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<<variable_expression.ast_twin, "_address">>, node.type_binding))
					declaration.needs_address := True
				elseif
					-- handle taking addresses of struct fields
					attached {AST_POSTFIX_EXPRESSION} node.expression as pe and then
					(pe.type = C_op_point or pe.type = C_op_arrow)
				then
					handle_struct_field_address (node)
				end
			end
		end

feature {NONE}

	handle_struct_field_address (node: AST_UNARY_EXPRESSION)
			-- `&a->b->c.d' -> `create {CE_POINTER [?]}.make_shared (a->b.c_pointer + {struct ?}.c_offset + {struct ?}.d_offset)'
		require
			node /= Void and then node.type = C_op_bitand
			attached {AST_POSTFIX_EXPRESSION} node.expression as pe and then (pe.type = C_op_point or pe.type = C_op_arrow or pe.type = C_op_open_square_bracket)
		local
			base: AST_EXPRESSION
			fields: TWO_WAY_LIST [AST_POSTFIX_EXPRESSION]
			cur_expr: AST_EXPRESSION
			a: ARRAYED_LIST [ANY]
			field: AST_VARIABLE_DECLARATION
			declaring_struct_type: AST_STRUCT_TYPE
			cur_type: AST_TYPE
			l_sizeof_expression: AST_SIZEOF_EXPRESSION
			l_expression_statement: AST_EXPRESSION_STATEMENT
		do
			-- disassemble access into `base' and `fields'
			from
				create fields.make
				cur_expr := node.expression
			until
				cur_expr = Void
			loop
				if attached {AST_POSTFIX_EXPRESSION} cur_expr as pe then
					fields.put_front (pe)
					inspect
						pe.type
					when C_op_point, C_op_open_square_bracket then
						cur_expr := pe.main_expression
					when C_op_arrow then
						base := pe.main_expression
						cur_expr := Void
					end
				else
					base := cur_expr
					cur_expr := Void
				end
			end

			-- create eiffel expression array
			create a.make (50)
			a.extend ("create {")
			a.extend (node.type_binding.ast_twin)
			a.extend ("}.make_shared (")
			a.extend (base.ast_twin)
			across fields as c loop
				inspect
					c.item.type
				when C_op_point, C_op_arrow then
					field ?= c.item.declaration_binding -- can't take address of bit fields
					inspect
						c.item.type
					when C_op_point then
						declaring_struct_type ?= c.item.main_expression.type_binding
					when C_op_arrow then
						declaring_struct_type ?= ({AST_POINTER_TYPE} #? c.item.main_expression.type_binding).type
					end

					a.extend (" + (create {")
					a.extend (declaring_struct_type.ast_twin)
					a.extend ("}.make).offset_of (%"")
					a.extend (create {AST_VARIABLE_EXPRESSION}.make (field.identifier, field))
					a.extend ("%")")
				when C_op_open_square_bracket then
					a.extend (" + ")
					 -- `x[a][b][c]' -> `sizeof(x) * (a * dim_b * dim_c + b * dim_c + c)'
					if not (attached {AST_POSTFIX_EXPRESSION} c.item.main_expression as pe and then pe.type = C_op_open_square_bracket) then
						-- `sizeof(x) * ('
						create l_sizeof_expression.make (({AST_ARRAY_TYPE} #? c.item.main_expression.type_binding).non_array_type.ast_twin)
						create l_expression_statement.make (l_sizeof_expression)
						l_sizeof_expression.accept (create {SIZEOF_TRANSFORMER})
						a.extend (l_expression_statement.expression.ast_twin)
						({STRING_8} #? ({AST_EIFFEL_EXPRESSION} #? a.last).mixed.last).remove_tail (14) -- remove ".to_natural_32"
						a.extend (" * (")
					end
					-- `a * dim_b * dim_c'
					a.extend (c.item.array_expression.ast_twin)
					from
						cur_type := c.item.type_binding
					until
						not (attached {AST_ARRAY_TYPE} cur_type as at)
					loop
						a.extend (" * ")
						a.extend (at.dimension.ast_twin)
						cur_type := at.type
					end
					if not (attached {AST_POSTFIX_EXPRESSION} c.item.parent as pe and then pe.type = C_op_open_square_bracket) then
						-- `)'
						a.extend (")")
					end
				end
			end
			a.extend (")")
			({STRING} #? a[5]).prepend (".c_pointer")

			replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (a.to_array, {AST_TYPE} #? a[2]))
		end

end
