class
	LITERAL_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_string_constant,
			end_visit_character_constant,
			end_visit_floating_point_constant,
			end_visit_hex_constant
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING_8
		do
			Result := "Transforming literals"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_string_constant (node: AST_STRING_CONSTANT)
		do
			transform_text_literal (node.value)
		end

	end_visit_character_constant (node: AST_CHARACTER_CONSTANT)
		do
			transform_text_literal (node.value)
		end

	end_visit_floating_point_constant (node: AST_FLOATING_POINT_CONSTANT)
		do
			-- "2e-3" -> "2.e-3"
			if (node.value.has ('e') or node.value.has ('E')) and not node.value.has ('.') then
				node.value.insert_character ('.', node.value.as_lower.index_of ('e', 1))
			end
		end

	end_visit_hex_constant (node: AST_HEX_CONSTANT)
		local
			l_function_name: STRING_8
		do
			-- `0x1.fp1' -> `hex_to_real_64 ("0x1.fp1")'
			inspect node.type_binding.type
			when Ast_bt_float       then l_function_name := "hex_to_real_32"
			when Ast_bt_double      then l_function_name := "hex_to_real_64"
			when Ast_bt_long_double then l_function_name := "hex_to_real_96"
			else
				-- no floating point number
			end

			if l_function_name /= Void then
				replace_node (node, create {AST_EIFFEL_EXPRESSION}.make (l_function_name + " (%"" + node.value + "%")", node.type_binding))
			end
		end

feature {NONE}

	transform_text_literal (s: STRING)
		require
			s /= Void
		local
			buf: STRING
			i, tmp: INTEGER_32
		do
			create buf.make_empty
			from
				i := 1
			until
				i > s.count
			loop
				if s[i] = '%%' then
					buf.append ("%%%%")
				elseif s[i] = '\' then
					inspect
						s[i + 1]
					when '%"' then buf.append ("%%%"")
					when '%'' then buf.append ("%%%'")
					when '\' then  buf.append ("\")
					when 'a' then  buf.append ("%%/7/") -- %A doesn't work
					when 'b' then  buf.append ("%%B")
					when 'f' then  buf.append ("%%F")
					when 'n' then  buf.append ("%%N")
					when 'r' then  buf.append ("%%R")
					when 't' then  buf.append ("%%T")
					when 'v' then  buf.append ("%%/11/") -- %V doesn't work
					when '0'..'7' then
						-- Transform octal value to integer since Eiffel doesn't
						-- support hex and octal constants in strings (PR #15042).
						tmp := s[i + 1].code - 48
						if s[i + 2] >= '0' and s[i + 2] <= '7' then
							tmp := tmp * 8 + s[i + 2].code - 48
							if s[i + 3] >= '0' and s[i + 3] <= '7' then
								tmp := tmp * 8 + s[i + 3].code - 48
								i := i + 1
							end
							i := i + 1
						end
						buf.append ("%%/" + tmp.out + "/")
					else
						buf.append_character (s[i+1])
					end
					i := i + 1
				else
					buf.append_character (s[i])
				end
				i := i + 1
			end
			s.make_from_string (buf)
		end

end
