class
	OO_INHERITANCE_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_eiffel_expression
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		local
			l_s: STRING_8
		do
			--transformer.tree.accept (Current)
			Result := True

			across transformer.oo_cluster.classes as c loop
				if not c.item.superclasses.is_empty then
					l_s := c.item.name + " inherits: "
					across c.item.superclasses as c2 loop
						if c2.target_index > 1 then
							l_s.append (", ")
						end
						l_s.append (c2.item.name)
					end
					log (l_s)
				end
			end
		end

	description: STRING
		do
			Result := "Inferring O-O inheritance structure"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_eiffel_expression (node: AST_EIFFEL_EXPRESSION)
		local
			l_subclass, l_superclass: OO_CLASS
		do
			if
				node.mixed.count = 5 and then
				attached {AST_STRUCT_DECLARATION} pointer_to_struct (node.mixed.i_th (2)) as dest_struct and then
				node.is_string (3, "}.make_cast (") and then
				attached {AST_STRUCT_DECLARATION} pointer_to_struct (node.mixed.i_th (4)) as src_struct and then
				dest_struct.is_union = src_struct.is_union and then
				dest_struct.fields.count /= src_struct.fields.count and then
				equal_fields (dest_struct, src_struct)
			then
				across
					transformer.oo_cluster.classes as c
				until
					l_subclass /= Void and l_superclass /= Void
				loop
					if attached {OO_DATA_CLASS} c.item as dc then
						if dc.node = dest_struct or dc.node = src_struct then
							if
								dc.node.fields.count > dest_struct.fields.count or
								dc.node.fields.count > src_struct.fields.count
							then
								l_subclass := dc
							else
								l_superclass := dc
							end
						end
					end
				end
				l_subclass.superclasses.extend (l_superclass)
			end
		end

feature {NONE}

	pointer_to_struct (a_node: ANY): AST_STRUCT_DECLARATION
		require
			a_node /= Void
		do
			if attached {AST_EXPRESSION} a_node as e then
				Result := pointer_to_struct (e.type_binding)
			elseif
				attached {AST_POINTER_TYPE} a_node as pt and then
				attached {AST_STRUCT_TYPE} pt.type as st
			then
				Result := st.declaration_binding
			elseif
				attached {AST_ARRAY_TYPE} a_node as at and then
				attached {AST_STRUCT_TYPE} at.type as st
			then
				Result := st.declaration_binding
			end
		end

	equal_fields (a_struct1, a_struct2: AST_STRUCT_DECLARATION): BOOLEAN
		require
			a_struct1 /= Void
			a_struct2 /= Void
		local
			l_struct1_cursor, l_struct2_cursor: CURSOR
		do
			l_struct1_cursor := a_struct1.fields.cursor
			l_struct2_cursor := a_struct2.fields.cursor
			from
				Result := True
				a_struct1.fields.start
				a_struct2.fields.start
			until
				Result = False or
				a_struct1.fields.after or
				a_struct2.fields.after
			loop
				Result := a_struct1.fields.item.type.is_equal_type (a_struct2.fields.item.type)
				a_struct1.fields.forth
				a_struct2.fields.forth
			end
			a_struct1.fields.go_to (l_struct1_cursor)
			a_struct2.fields.go_to (l_struct2_cursor)
		end

end
