class
	DEPENDENCY_GRAPH_ITEM [G]

inherit
	ANY
		redefine
			is_equal
		end
	DEPENDENCY_GRAPH_FLAGS
		undefine
			is_equal
		end

create
	make

feature

	make (a_key: G; an_identifier: STRING; a_flags: like flags)
		do
			key := a_key
			identifier := an_identifier
			is_bad := False
			flags := a_flags
			reference_number := 0
			create dependencies.make

			if is_extern then
				-- ensure invariant
				mark_bad
			end
		end

feature

	key: G

	identifier: STRING assign set_identifier

	bad_marks: INTEGER

	flags: INTEGER_32

	is_bad: BOOLEAN

	is_extern: BOOLEAN
		do
			Result := has_flag (flags, external_flag)
		end

feature {DEPENDENCY_GRAPH, DEPENDENCY_GRAPH_ITEM}

	dependencies: LINKED_LIST [like Current]

	reference_number: INTEGER assign set_ref_number

	set_ref_number (n: like reference_number)
		do
			reference_number := n
		end

	set_identifier (an_identifier: like identifier)
		do
			identifier := an_identifier
		end

	add_flag (new_flag: like flags)
		do
			if new_flag /= 0 then
				flags := flags | new_flag
				mark_bad
			end
		end

feature

	mark_bad
		do
			is_bad := True
			bad_marks := bad_marks + 1
		end

	add_dep (dep: like Current)
		do
			if not dependencies.has (dep) then
				dependencies.extend (dep)
			end
		end

	to_string: STRING
		do
			create Result.make_from_string (reference_number.out + ": " + identifier)

			if not is_bad then
				Result.append (" (+++++)")
			else
				Result.append (" [" + bad_marks.out + ", " + flags_to_string (flags) + "]")
			end

			if not dependencies.is_empty then
				Result.append (" ->")
				across
					dependencies as d
				loop
					Result.append (" " + d.item.reference_number.out)
				end
			end
		end

	is_equal (other: like Current): BOOLEAN
		do
			Result := key.is_equal (other.key)
		end

invariant
	is_extern implies is_bad

end
